from painy.git import *
from painy.chat import Model
from painy import console
from painy.utils import print_commit_message, load_json, load_txt 
from painy.errors import NoSettingException
from painy.managers import ConfigManager, RulesManager


def get_rules(config: dict):
    rules_loaded = load_json("settings/rules.json")['rules']
    rules_processed = process_rules(rules_loaded, config)
    
    rules_str = '\n\nFollow this rules/restrictions when writing commit messages:\n'
    rules_str += '\n'.join([f'- {rule.strip()}' for rule in rules_processed])
            
    return rules_str
   
 
def get_prompt() -> str:
    prompt = load_txt('settings/prompt_comment_base.txt')
    config_manager = ConfigManager()
    rules_manager = RulesManager()
    rules = rules_manager.get_rules(config_manager.config_dict)
    
    rules_str = '\n\nFollow this rules/restrictions when writing commit messages:\n'
    rules_str += '\n'.join([f'- {rule.strip()}' for rule in rules])
    
    if config_manager.get_option('use_commit_history_style'):
        max_num_commits = config_manager.get_option('max_num_commits_style')
        max_num_commits = max_num_commits if max_num_commits is not None else 20
        commit_history = get_commit_messsage_history()[:max_num_commits]
        
        style_prompt = load_txt('settings/prompt_commit_style.txt')
        prompt += '\n\n' + style_prompt + '\n'
        prompt += '\n'.join([f'- {rule.strip()}' for rule in commit_history])
    
    return prompt


def get_commmit_message(diff_str: str, interactive: bool = False):
    with console.status(status="Generating commit message...", spinner='aesthetic'):
        comment_prompt = get_prompt()
        model = Model(purpose_prompt=comment_prompt)
        response = model.get_response(prompt=diff_str)
    
    return response


def comment_interactive(msg: str, diff_str: str) -> str:
    """
    Interactive mode for generating commit messages.
    
    Args:
        msg (str): Already (first time) generated commit message
    
    Returns:
        str: Final commit message
    """
    while True:
        console.rule(characters="=", style="cyan")
        option = console.input("Try another one? [green]y[/green]/[red]n[/red]: ")
        
        if option.lower() in ["y", "yes"]:
            msg = get_commmit_message(diff_str)
            print_commit_message(console, msg)
        else:
            return msg


import os
import rich
import json


def get_package_path() -> str:
    return os.path.dirname(os.path.realpath(__file__))

def load_valid_extensions() -> str:
    with open(os.path.join(get_package_path(), "settings/p_extensions.txt")) as f:
        extensions = f.read().splitlines()
    
    return extensions

def load_json(path: str) -> dict:
    with open(os.path.join(get_package_path(), path)) as f:
        config = json.load(f)
    
    return config

def save_json(path: str, data: dict) -> None:
    with open(os.path.join(get_package_path(), path), 'w') as f:
        json.dump(data, f, indent=4)

def load_txt(path: str) -> str:
    with open(os.path.join(get_package_path(), path)) as f:
        prompt = f.read()
    
    return prompt

def print_commit_message(console: rich.console.Console, msg: str) -> None:
    console.print(f"[bold]Commit message[/bold]:\n[green]{msg}[/green]")


from painy.utils import load_json, save_json
import re


class ConfigManager:
    def __init__(self, config_path='settings/config.json'):
        self.config_path = config_path
        self.config_dict = load_json(self.config_path)
        
    def set_option(self, option: str, value) -> None:
        self.config_dict[option] = value
        self.save_config()
        
    def get_option(self, option: str):
        if option not in self.config_dict:
            return None
        value = self.config_dict[option]
        return ConfigManager.process(value)
        
    def save_config(self) -> None:
        save_json(self.config_path, self.config_dict)
    
    @staticmethod
    def process(value):
        if type(value) == str:
            if value.lower() == 'true':
                return True
            elif value.lower() == 'false':
                return False
            elif value.isdigit():
                return int(value)

        return value
       
        
class RulesManager:
    def __init__(self, rules_path='settings/rules.json'):
        self.rules_path = rules_path
        self.rules_dict = load_json(self.rules_path)
        
    def add_rule(self, rule: str) -> None:
        self.rules_dict['rules'] = self.rules_dict['rules'] + [rule]
        self.save_rules()
        
    def remove_rule(self, i) -> None:
        self.rules_dict['rules'].pop(i)
        self.save_rules()
        
    def get_rules(self, config: dict = None) :
        rules_loaded = self.rules_dict['rules']
        if config is None:
            return rules_loaded
        return RulesManager.process_rules(rules_loaded, config)
    
    def get_length(self) -> int:
        return len(self.rules_dict['rules'])
    
    def remove_all_rules(self) -> None:
        self.rules_dict['rules'] = []
        self.save_rules()
    
    def save_rules(self) -> None:
        save_json(self.rules_path, self.rules_dict)
        
    @staticmethod
    def process_rules(rules: list, config: dict) -> str:
        pattern = r"(\<.+\>)"
        
        processed = []
        for rule in rules:
            var_matches = re.findall(pattern, rule)
            if not var_matches:
                processed.append(rule)
            for match in var_matches:
                var = match.strip('<>')
                
                if var in config:
                    rule = rule.replace(match, str(config[var]))
                    processed.append(rule)
                else:
                    raise NoSettingException(f"Setting {var} not found in config.json")
            
        return processed


from .git import *
from .enums import *
import os
from langchain_community.chat_models import ChatOpenAI
from langchain_core.runnables import RunnableLambda, RunnablePassthrough
from langchain.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain.chains import ConversationChain
from langchain_core.messages import SystemMessage
from operator import itemgetter
from langchain.memory import ConversationBufferMemory


MODEL_NAME = os.getenv(key="OPENAI_MODEL_NAME", default="gpt-3.5-turbo")


class Model:
    def __init__(self, purpose_prompt: str, model_name: str = MODEL_NAME):
        self.model_name = MODEL_NAME

        self.llm = ChatOpenAI(
            temperature=0.1,
            model_name=self.model_name
        )

        self.memory = ConversationBufferMemory(return_messages=True)
        self.memory.chat_memory.add_message(
            SystemMessage(content=purpose_prompt))

        self.chain = ConversationChain(
            llm=self.llm,
            verbose=False,
            memory=self.memory
        )

    def get_response(self, prompt: str) -> str:
        inputs = {"input": prompt}
        response = self.chain.run(inputs)

        return response



class GitDiffException(Exception):
    def __init__(self, message: str = "Error. Are you in a git repository?"):
        self.message = message
        super().__init__(self.message)
    def __str__(self):
        return self.message


from typing import List
import subprocess
import re
from painy.utils import get_package_path, load_valid_extensions
from painy.errors import *


def get_changed_files(staged=False) -> List[str]:
    try:
        cmd = ["git", "diff", "--name-only"]
        
        if staged:
            cmd += ["--staged"]
        
        output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        
        changed_files = output.decode("utf-8").strip().split("\n")
    except subprocess.CalledProcessError:
        raise GitDiffException()
    
    valid_extensions = load_valid_extensions()
     
    changed_files = [file for file in changed_files if file != '']
    changed_files = [file for file in changed_files if f".{file.split('.')[-1]}" in valid_extensions]
    
    return changed_files

def get_ipynb_changes_staged(file_path: str) -> str:
    assert file_path.endswith(".ipynb")
    
    cmd = ["nbdiff", file_path]    
    output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
    changes = output.decode("utf-8").strip() + "\n"
    
    # Remove additional info
    regs = [
        "##.* added \/nbformat:[\n \S]*?\n\n",
        "##.* added \/nbformat_minor:[\n \S]*?\n\n",
        "##.* added \/metadata:[\n \S]*?\n\n"]
    
    for reg in regs:
        changes = re.sub(reg, "", changes)
    
    return changes

def get_file_changes(file_path: str) -> str:
    if file_path.endswith(".ipynb"):
        changes = get_ipynb_changes_staged(file_path)
    else:
        output = subprocess.check_output(["git", "diff", "HEAD", "--", file_path])
        changes = output.decode("utf-8")
    return changes

def get_changes_staged() -> str:
    output = subprocess.check_output(["git", "diff", "--staged"], stderr=subprocess.STDOUT)
    changes = output.decode("utf-8").strip()
    return changes

def get_diff_str(changed_files: List[str]) -> str:
    if len(changed_files) == 0:
        raise NoChangesException()
    
    diffs = []
    
    for file in changed_files:
        diff = get_file_changes(file)
        diffs.append(diff)
    
    return "\n".join(diffs)

def get_commit_messsage_history() -> List[str]:
    output = subprocess.check_output(["git", "log", "--pretty=format:\"%s\""], stderr=subprocess.STDOUT)
    messages = output.decode("utf-8").strip().split("\n")
    messages = [message.strip("\"") for message in messages]
    
    return messages

def commit(commit_message: str) -> None:
    """
    Commits the staged changes with the given commit message.
    Used to commit the auto-generated commit message.
    
    Args:
        commit_message (str): The commit message to use.
    """
    subprocess.check_output(["git", "commit", "-m", commit_message], stderr=subprocess.STDOUT)


from setuptools import find_packages, setup

with open('requirements.txt') as f:
    requirements = f.readlines()

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="Painy",
    version="0.2.1",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Raevskiy Rudolf",
    url="https://github.com/lanesket/painy",
    packages=find_packages(),
    python_requires=">=3.7.1",
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    package_data={
        'painy': ['settings/*']
    },
    entry_points={
        'console_scripts': [
            "painy = painy.main:main"
        ]
    }
)


import argparse
import openai
import os
from painy import console
from painy.comment import get_commmit_message, comment_interactive
from painy.enums import Action
from painy.git import commit, get_changed_files, get_diff_str
from painy.utils import print_commit_message
from painy.managers import ConfigManager, RulesManager


def main():
    api_key = os.getenv("OPENAI_API_KEY")

    if api_key is None:
        console.print("[red]Error: OPENAI_API_KEY not set[/red]")
        exit(1)

    openai.api_key = api_key
    
    parser = argparse.ArgumentParser(
        prog="Painy",
        description="A tool to help you write better commit messages."
    )
    
    parser.add_argument("action", choices=["comment", "commit", "config", "rules"], help="The action to perform")
    parser.add_argument("--check-all", action="store_true", help="Check all files previously registered in git, not just staged ones")
    parser.add_argument("-i", "--interactive", action="store_true", help="Run in interactive mode")
    parser.add_argument("-s", "--set", nargs=2, help="Set a config value. Example: painy config --set use_commit_history_style True")
    parser.add_argument("-g", "--get", help="Get a config value. Example: painy config --get use_commit_history_style")
    parser.add_argument("-a", "--add", help="Add a rule. Example: painy rules --add '<string>'")
    parser.add_argument("-r", "--remove", help="Remove a i-th rule from the list. Example: painy rules --remove <int>")
    
    args = parser.parse_args()
    
    action = args.action if args.action is not None else Action.COMMENT.value
    
    staged = not args.check_all
    interactive = args.interactive
    
    if action == Action.CONFIG.value:
        if args.set is not None and len(args.set) == 2:
            key, value = args.set
            console.print(f"[green]Setting config value:[/green] {key} = {value}")
                        
            config_manager = ConfigManager()
            config_manager.set_option(key, value)
        elif args.get is not None:
            key = args.get
            console.print(f"[green]Getting config value:[/green] {key}")
            
            config_manager = ConfigManager()
            value = config_manager.get_option(key)
            
            console.print(f"[green]Value:[/green] {value}") 
    
        return
    
    if action == Action.RULES.value:
        rules_manager = RulesManager()
        config_manager = ConfigManager()

        if args.add is not None:
            rule = args.add
            console.print(f"[green]Adding rule:[/green] {rule}")
            
            rules_manager.add_rule(rule)
        elif args.remove is not None:
            i = int(args.remove)
            if i <= 0:
                console.print("[red]Error: i must be greater than 0[/red]")
                return
            
            console.print(f"[green]Removing rule:[/green] {i}")
            try:
                rules_manager.remove_rule(i - 1)
            except IndexError:
                length = rules_manager.get_length()
                console.print(f"[red]Error: i must be less than or equal to {length}[/red]")
        else:
            rules = rules_manager.get_rules(config_manager.config_dict)
            rules_str = '\n'.join(rules)
            console.print(f"[green]Rules:[/green]\n{rules_str}")
        
        return
    
    try:
        changed_files = get_changed_files(staged)
        console.print(f"[green]Changed files:[/green] {', '.join(changed_files)}")
        
        diff_str = get_diff_str(changed_files)
    except Exception as e:
        console.print(f"[red]{e}[/red]")
        return
    
    if action == Action.COMMENT.value:
        msg = get_commmit_message(diff_str)
        print_commit_message(console, msg)
        
        if interactive:
            msg = comment_interactive(msg, diff_str)
            
    elif action == Action.COMMIT.value:
        msg = get_commmit_message(diff_str)
        print_commit_message(console, msg)
         
        if interactive:
            msg = comment_interactive(msg, diff_str)
            
            option = console.input("Do you want to commit with this message? [green]y[/green]/[red]n[/red]: ")
            
            if option.lower() in ["y", "yes"]:
                commit(msg)
        else:
            print_commit_message(console, msg)
            commit(msg)
        

if __name__ == "__main__":
    main()